from sqlalchemy import Integer, Unicode

from restfulpy.orm import DeclarativeBase, Field, ActivationMixin, \
    AutoActivationMixin


class ActiveObject(ActivationMixin, DeclarativeBase):
    __tablename__ = 'active_object'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))


class AutoActiveObject(AutoActivationMixin, DeclarativeBase):
    __tablename__ = 'auto_active_object'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))


def test_activation_mixin(db):
    session = db(expire_on_commit=False)

    object1 = ActiveObject(
        title='object 1',
    )

    session.add(object1)
    session.commit()
    assert not object1.is_active
    assert session.query(ActiveObject)\
        .filter(ActiveObject.is_active)\
        .count() == 0

    object1.is_active = True
    assert object1.is_active
    session.commit()

    object1 = session.query(ActiveObject).one()
    assert object1.is_active

    assert 'isActive' in object1.to_dict()

    assert session.query(ActiveObject)\
        .filter(ActiveObject.is_active)\
        .count() == 1
    assert ActiveObject.filter_activated(session.query(ActiveObject)).count()\
        == 1

    assert not ActiveObject.import_value(ActiveObject.is_active, 'false')
    assert not ActiveObject.import_value(ActiveObject.is_active, 'FALSE')
    assert not ActiveObject.import_value(ActiveObject.is_active, 'False')
    assert ActiveObject.import_value(ActiveObject.is_active, 'true')
    assert ActiveObject.import_value(ActiveObject.is_active, 'TRUE')
    assert ActiveObject.import_value(ActiveObject.is_active, 'True')
    assert ActiveObject.import_value(ActiveObject.title, 'title') == 'title'


def test_auto_activation(db):
    session = db()
    object1 = AutoActiveObject(
        title='object 1',
    )

    session.add(object1)
    session.commit()
    assert object1.is_active
    assert session.query(AutoActiveObject)\
        .filter(AutoActiveObject.is_active)\
        .count() == 1


def test_metadata():
    # Metadata
    object_metadata = ActiveObject.json_metadata()
    fields = object_metadata['fields']

    assert 'id' in fields
    assert 'title' in fields
    assert 'isActive' in fields




import time

from bddrest import status, response, when
from freezegun import freeze_time
from nanohttp import json, Controller, context
from nanohttp.contexts import Context

from restfulpy.mockup import MockupApplication
from restfulpy.authentication import StatefulAuthenticator
from restfulpy.authorization import authorize
from restfulpy.principal import JwtPrincipal, JwtRefreshToken
from restfulpy.testing import ApplicableTestCase


class MockupMember:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


roles = ['admin', 'test']


class MockupStatefulAuthenticator(StatefulAuthenticator):
    def validate_credentials(self, credentials):
        email, password = credentials
        if password == 'test':
            return MockupMember(id=1, email=email, roles=['admin', 'test'])

    def create_refresh_principal(self, member_id=None):
        return JwtRefreshToken(dict(
            id=member_id
        ))

    def create_principal(self, member_id=None, session_id=None):
        return JwtPrincipal(dict(
            id=1,
            email='test@example.com',
            roles=roles,
            sessionId='1'
        ))


class Root(Controller):
    @json
    def login(self):
        principal = context.application.__authenticator__.login(
            (context.form['email'], context.form['password'])
        )
        return dict(token=principal.dump())

    @json
    @authorize
    def me(self):
        return context.identity.payload

    @json
    @authorize
    def invalidate_token(self):
        context.application.__authenticator__.invalidate_member(1)
        return context.identity.payload

    @json(verbs='delete')
    @authorize
    def logout(self):
        context.application.__authenticator__.logout()
        return {}


class TestStatefulAuthenticator(ApplicableTestCase):
    __application__ = MockupApplication(
        'Stateful Authenticator Application',
        Root(),
        authenticator=MockupStatefulAuthenticator()
    )

    __configuration__ = '''
        jwt:
          max_age: .3
          refresh_token:
            max_age: 3
            secure: false
    '''

    def test_invalidate_token(self):
        with self.given(
                'Log in to get a token and refresh token cookie',
                '/login',
                'POST',
                form=dict(email='test@example.com', password='test')
            ):
            assert status == 200
            assert 'token' in response.json
            refresh_token = response.headers['Set-Cookie'].split('; ')[0]
            assert refresh_token.startswith('refresh-token=')

            # Store token to use it for future requests
            token = response.json['token']
            self._authentication_token = token

        with self.given(
                'Request a protected resource to ensure authenticator is '
                'working well',
                '/me',
                headers={'Cookie': refresh_token}
            ):
            assert status == 200
            assert response.json['roles'] == roles

            roles.append('god')
            when(
                'Invalidating the token by server',
                '/invalidate_token'
            )
            assert response.json['roles'] == roles
            assert 'X-New-JWT-Token' in response.headers

            when(
                'Invalidating the token by server after the token has '
                'been expired, with appropriate cookies.',
                '/invalidate_token',
            )
            time.sleep(1)
            assert 'X-New-JWT-Token' in response.headers
            assert response.headers['X-New-JWT-Token'] is not None
            self._authentication_token = response.headers['X-New-JWT-Token']

            when(
                'Requesting a resource with new token',
                '/me',
            )
            assert status == 200

    def test_logout(self):
        with self.given(
                'Log in to get a token and refresh token cookie',
                '/login',
                'POST',
                form=dict(email='test@example.com', password='test')
            ):
            assert status == 200
            assert 'token' in response.json
            assert response.headers['X-Identity'] == '1'
            self._authentication_token = response.json['token']

            when(
                'Logging out',
                '/logout',
                'DELETE'
            )
            assert 'X-Identity' not in response.headers

    def test_session_member(self):
        with Context(environ={}, application=self.__application__):
            authenticator = self.__application__.__authenticator__
            principal = authenticator.login(
                ('test@example.com', 'test')
            )
            assert authenticator.get_member_id_by_session(
                principal.session_id
            ) == 1

    @freeze_time("2017-07-13T13:11:44", tz_offset=-4)
    def test_session_info(self):
        with self.given(
                'Log in to get a token and refresh token cookie',
                '/login',
                'POST',
                form=dict(email='test@example.com', password='test')
            ):
            assert status == 200
            assert 'token' in response.json
            assert response.headers['X-Identity'] == '1'
            self._authentication_token = response.json['token']

            # Testing test cases
            for test_case in session_info_test_cases:
                # Our new session info should be updated
                self.when(
                    'Getting session info',
                    '/me',
                    extra_environ=test_case['environment']
                )
                assert status == 200
                assert 'sessionId' in response.json

                info = self.__application__.__authenticator__\
                    .get_session_info(response.json['sessionId'])

                assert info.items() == {
                    'remoteAddress': test_case['expected_remote_address'],
                    'machine': test_case['expected_machine'],
                    'os': test_case['expected_os'],
                    'agent': test_case['expected_agent'],
                    'client': test_case['expected_client'],
                    'app': test_case['expected_app'],
                    'lastActivity': test_case['expected_last_activity'],
                }.items()


session_info_test_cases = [
    {
        'environment': {
            'REMOTE_ADDR': '',
            'HTTP_USER_AGENT': \
                'Mozilla/5.0 (iPhone; CPU iPhone OS 5_1 like Mac OS X) '
                'AppleWebKit/534.46 (KHTML, like Gecko) Version/5.1 '
                'Mobile/9B179 Safari/7534.48.3 RestfulpyClient-js/1.2.3 (My '
                'App; test-name; 1.4.5-beta78; fa-IR; some; extra; info)'
        },
        'expected_remote_address': 'NA',
        'expected_machine': 'iPhone',
        'expected_os': 'iOS 5.1',
        'expected_agent': 'Mobile Safari 5.1',
        'expected_client': 'RestfulpyClient-js 1.2.3',
        'expected_app': 'My App (test-name) 1.4.5-beta78',
        'expected_last_activity': '2017-07-13T13:11:44',
    },
    {
        'environment': {
            'REMOTE_ADDR': '185.87.34.23',
            'HTTP_USER_AGENT': \
                'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; '
                'Trident/5.0) RestfulpyClient-custom/4.5.6 (A; B; C)'
        },
        'expected_remote_address': '185.87.34.23',
        'expected_machine': 'PC',
        'expected_os': 'Windows 7',
        'expected_agent': 'IE 9.0',
        'expected_client': 'RestfulpyClient-custom 4.5.6',
        'expected_app': 'A (B) C',
        'expected_last_activity': '2017-07-13T13:11:44',
    },
    {
        'environment': {
            'REMOTE_ADDR': '172.16.0.111',
            'HTTP_USER_AGENT': ''
        },
        'expected_remote_address': '172.16.0.111',
        'expected_machine': 'Other',
        'expected_os': 'Other',
        'expected_agent': 'Other',
        'expected_client': 'Unknown',
        'expected_app': 'Unknown',
        'expected_last_activity': '2017-07-13T13:11:44',
    },
    {
        'environment': {},
        'expected_remote_address': 'NA',
        'expected_machine': 'Other',
        'expected_os': 'Other',
        'expected_agent': 'Other',
        'expected_client': 'Unknown',
        'expected_app': 'Unknown',
        'expected_last_activity': '2017-07-13T13:11:44',
    }
]



import asyncore
import contextlib
import io
import smtpd
import threading
from http.server import BaseHTTPRequestHandler, HTTPStatus, HTTPServer
from wsgiref.simple_server import WSGIServer, WSGIRequestHandler

from . import datetimehelpers
from .messaging import Messenger
from .mimetypes_ import guess_type
from .utils import copy_stream
from .application import Application


SERVER_LOCK = threading.Event()


@contextlib.contextmanager
def mockup_http_server(app=None, handler_class=WSGIRequestHandler,
                server_class=WSGIServer, bind=('', 0)):
    server = server_class(bind, handler_class)
    if app:
        assert isinstance(server, WSGIServer)
        server.set_app(app)
    thread = threading.Thread(
        target=server.serve_forever,
        name='sa-media test server.',
        daemon=True
    )
    thread.start()
    url = 'http://localhost:%s' % server.server_address[1]
    yield server, url
    server.shutdown()
    thread.join()


def mockup_http_static_server(content: bytes = b'Simple file content.',
                              content_type: str = None, **kwargs):
    class StaticMockupHandler(BaseHTTPRequestHandler):  # pragma: no cover
        def serve_text(self):
            self.send_header('Content-Type', "text/plain")
            self.send_header('Content-Length', str(len(content)))
            self.send_header('Last-Modified', self.date_time_string())
            self.end_headers()
            self.wfile.write(content)

        def serve_static_file(self, filename):
            self.send_header('Content-Type', guess_type(filename))
            with open(filename, 'rb') as f:
                self.serve_stream(f)

        def serve_stream(self, stream):
            buffer = io.BytesIO()
            self.send_header(
                'Content-Length',
                str(copy_stream(stream, buffer))
            )
            self.end_headers()
            buffer.seek(0)
            try:
                copy_stream(buffer, self.wfile)
            except ConnectionResetError:
                pass

        def do_GET(self):
            self.send_response(HTTPStatus.OK)
            if isinstance(content, bytes):
                self.serve_text()
            elif isinstance(content, str):
                self.serve_static_file(content)
            else:
                self.send_header('Content-Type', content_type)
                self.serve_stream(content)

    return simple_http_server(
        None,
        handler_class=StaticMockupHandler,
        server_class=HTTPServer,
        **kwargs
    )


class MockupSMTPServer(smtpd.SMTPServer):

    def __init__(self, bind):
        super().__init__(bind, None, decode_data=False)
        self.server_address = self.socket.getsockname()[:2]
        SERVER_LOCK.set()

    def process_message(*args, **kwargs):
        pass


@contextlib.contextmanager
def mockup_smtp_server(bind=('localhost', 0)):
    server = MockupSMTPServer(bind)
    thread = threading.Thread(target=asyncore.loop, daemon=True)
    thread.start()
    SERVER_LOCK.wait()
    yield server, server.server_address
    asyncore.close_all()


class MockupMessenger(Messenger):
    _last_message = None

    @property
    def last_message(self):
        return self.__class__._last_message

    @last_message.setter
    def last_message(self, value):
        self.__class__._last_message = value

    def send(
            self,
            to, subject, body,
            cc=None,
            bcc=None,
            template_string=None,
            template_filename=None,
            from_=None,
            attachments=None
    ):
        self.last_message = {
            'to': to,
            'body': body,
            'subject': subject
        }


@contextlib.contextmanager
def mockup_localtimezone(timezone):
    backup = datetimehelpers.localtimezone
    if callable(timezone):
        datetimehelpers.localtimezone = timezone
    else:
        datetimehelpers.localtimezone = lambda: timezone

    yield

    datetimehelpers.localtimezone = backup


class MockupApplication(Application):
    __configuration__ = '''
     db:
       url: postgresql://postgres:postgres@localhost/restfulpy_dev
       test_url: postgresql://postgres:postgres@localhost/restfulpy_test
       administrative_url: postgresql://postgres:postgres@localhost/postgres
    '''




from sqlalchemy import Integer, Unicode

from restfulpy.orm import DeclarativeBase, Field, DeactivationMixin


class DeactiveObject(DeactivationMixin, DeclarativeBase):
    __tablename__ = 'deactive_object'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))


def test_deactivation_mixin(db):
    session = db()

    object1 = DeactiveObject(
        title='object 1',
    )

    session.add(object1)
    session.commit()

    assert object1.is_active == False
    assert session.query(DeactiveObject).filter(DeactiveObject.is_active)\
        .count() == 0

    object1.is_active = True

    assert object1.is_active == True
    session.commit()
    object1 = session.query(DeactiveObject).one()
    assert object1.is_active == True
    assert object1.deactivated_at is None
    assert object1.activated_at is not None

    object1.is_active = False
    assert object1.is_active == False
    session.commit()
    object1 = session.query(DeactiveObject).one()

    assert object1.is_active == False
    assert object1.activated_at is None
    assert object1.deactivated_at is not None




import fcntl, termios, struct
import subprocess
from datetime import datetime


def terminal_size():
    h, w, hp, wp = struct.unpack(
        'HHHH',
        fcntl.ioctl(0, termios.TIOCGWINSZ, struct.pack('HHHH', 0, 0, 0, 0))
    )
    return w, h


class ProgressBar:

    def __init__(self, total):
        self._value = 0
        self.total = total
        self.start_time = None
        try:
            self.terminal_width = terminal_size()[0]
        except OSError:
            self.terminal_width = 120

    def increment(self):
        self._value += 1
        self._invalidate()

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v):
        self._value = v
        self._invalidate()

    @property
    def percent(self):
        return self._value * 100 // self.total

    @property
    def time(self):
        td = datetime.utcnow() - self.start_time
        return 'time: %.2d:%.2d' % (td.seconds // 60, td.seconds % 60)

    @property
    def estimated_time(self):
        if self.value is 0:
            et = 0
        else:
            td = datetime.utcnow() - self.start_time
            et = td.total_seconds() * (self.total / self.value - 1)
        return 'eta: %.2d:%.2d' % (int(et) // 60, int(et) % 60)

    @property
    def marks(self):
        scale = 2
        v = self.percent // scale
        return '%s%s' % ('#' * int(v), '.' * int(100 // scale - v))

    def get_progressbar_color(self):
        percent = self.percent
        if percent < 10:
            return red
        elif percent < 20:
            return lightred
        elif percent < 30:
            return yellow
        elif percent < 40:
            return lightyellow
        elif percent < 50:
            return violet
        elif percent < 60:
            return lightviolet
        elif percent < 70:
            return beige
        elif percent < 80:
            return lightbeige
        elif percent < 90:
            return blue
        elif percent < 100:
            return lightblue
        else:
            return green

    def _invalidate(self):
        detailed = \
            ('%%%dd/%%d' % len(str(self.total))) % (self._value, self.total)
        percent = '%3d%%' % self.percent
        progress = '|%s|' % self.marks
        line = ' '.join((
            detailed,
            self.get_progressbar_color(),
            percent,
            progress,
            clear,
            self.time,
            yellow,
            self.estimated_time,
            clear
        ))
        print(line, end='', flush=False)
        print(' ' * (self.terminal_width - len(line)), end='\r', flush=True)

    def __enter__(self):
        self.start_time = datetime.utcnow()
        self._invalidate()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._invalidate()
        print()


class LineReaderProgressBar(ProgressBar):
    """
    A proxy for IO file, with progressbar for reading file line by line.

    """
    def __init__(self, filename, mode='r'):
        self.filename = filename
        self.file = open(filename, mode)
        super().__init__(self.file_len(filename))

    @staticmethod
    def file_len(filename):
        p = subprocess.Popen(
            ['wc', '-l', filename],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        result, err = p.communicate()
        if p.returncode != 0:
            raise IOError(err)
        return int(result.strip().split()[0])

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.__exit__(exc_type, exc_val, exc_tb)
        return super().__exit__(exc_type, exc_val, exc_tb)

    def readline(self):
        self.increment()
        return self.file.readline()

    def __iter__(self):
        return self

    def __next__(self):
        self.increment()
        return self.file.__next__()


# default color
clear = '\33[22;24;25;27;28;39;49m'

# foreground colors:
black = '\33[30m'
red = '\33[31m'
green = '\33[32m'
yellow = '\33[33m'
blue = '\33[34m'
violet = '\33[35m'
beige = '\33[36m'
white = '\33[37m'
grey = '\33[90m'
gray = '\33[90m'
lightred = '\33[91m'
lightgreen = '\33[92m'
lightyellow = '\33[93m'
lightblue = '\33[94m'
lightviolet = '\33[95m'
lightbeige = '\33[96m'
lightwhite = '\33[97m'

# background colors:
bblack = '\33[40m'
bred = '\33[41m'
bgreen = '\33[4m'
byellow = '\33[43m'
bblue = '\33[44m'
bviolet = '\33[45m'
bbeige = '\33[46m'
bwhite = '\33[47m'
bgrey = '\33[100m'
bgray = '\33[100m'
blightred = '\33[101m'
blightgreen = '\33[102m'
blightyellow = '\33[103m'
blightblue = '\33[104m'
blightviolet = '\33[105m'
blightbeige = '\33[106m'
blightwhite = '\33[107m'

# styles
bold = '\33[1m'
italic = '\33[3m'



from datetime import datetime

import pytest
from bddrest import response, when, status
from dateutil.tz import tzoffset
from nanohttp import json
from sqlalchemy import Integer, Time

from restfulpy.configuration import settings
from restfulpy.controllers import JsonPatchControllerMixin, ModelRestController
from restfulpy.datetimehelpers import parse_datetime, format_datetime
from restfulpy.mockup import mockup_localtimezone
from restfulpy.orm import commit, DeclarativeBase, Field, DBSession
from restfulpy.testing import ApplicableTestCase


class Azan(DeclarativeBase):
    __tablename__ = 'azan'
    id = Field(Integer, primary_key=True)
    when = Field(Time)


class Root(JsonPatchControllerMixin, ModelRestController):
    __model__ = 'azan'

    @json
    @commit
    def post(self):
        m = Azan()
        m.update_from_request()
        DBSession.add(m)
        return m.when.isoformat()


class TestTime(ApplicableTestCase):
    __controller_factory__ = Root

    def test_update_from_request(self):
        with self.given(
            'Posting a time in form',
            verb='POST',
            form=dict(
                when='00:01',
            )
        ):
            assert status == 200
            assert response.json == '00:01:00'

            when(
                'Posting a date instead of time',
                form=dict(
                    when='2001-01-01'
                )
            )
            assert status == 200
            assert response.json == '00:00:00'

            when(
                'Posting a datetime instead of time',
                form=dict(
                    when='2001-01-01T00:01:00.123456'
                )
            )
            assert status == 200
            assert response.json == '00:01:00.123456'

            when(
                'Posting an invalid time',
                form=dict(
                    when=''
                )
            )
            assert status == '400 Invalid date or time: '

            when(
                'Posting another invalid time',
                form=dict(
                    when='invalid'
                )
            )
            assert status == '400 Invalid date or time: invalid'




import os
from os import path
import pytest
from bddrest import Given, when
from nanohttp import settings

from .mockup import MockupApplication
from .configuration import configure
from .db import DatabaseManager as DBManager
from .orm import setup_schema, session_factory, create_engine, init_model, \
    DBSession


@pytest.fixture(scope='module')
def db():
    builtin_configuration = '''
    db:
      test_url: postgresql://postgres:postgres@localhost/pytest_test
      administrative_url: postgresql://postgres:postgres@localhost/postgres
    logging:
      loggers:
        default:
          level: warning
    '''
    configure(builtin_configuration, force=True)

    # Overriding the db uri becase this is a test session, so db.test_uri will
    # be used instead of the db.uri
    settings.db.url = settings.db.test_url

    # Drop the previosely created db if exists.
    with DBManager() as m:
        m.drop_database()
        m.create_database()

    # An engine to create db schema and bind future created sessions
    engine = create_engine()

    # A session factory to create and store session to close it on tear down
    sessions = []
    def _connect(*a, expire_on_commit=False, **kw):
        new_session = session_factory(
            bind=engine,
            *a,
            expire_on_commit=expire_on_commit,
            **kw
        )
        sessions.append(new_session)
        return new_session

    session = _connect(expire_on_commit=True)

    # Creating database objects
    setup_schema(session)
    session.commit()

    # Closing the session to free the connection for future sessions.
    session.close()

    # Preparing and binding the application shared scoped session, due the
    # some errors when a model trying use the mentioned session internally.
    init_model(engine)

    yield _connect

    # Closing all sessions created by the test writer
    for s in sessions:
        s.close()
    DBSession.close_all()
    engine.dispose()

    # Dropping the previously created database
    with DBManager() as m:
        m.drop_database()


class TestCase:
    pass


class ApplicableTestCase:
    __application__ = None
    __application_factory__ = MockupApplication
    __controller_factory__ = None
    __configuration__ = None
    __story_directory__ = None
    __api_documentation_directory__ = None
    _engine = None
    _sessions = []
    _authentication_token = None

    @classmethod
    def configure_application(cls):
        cls.__application__.configure(force=True)
        settings.merge('''
          logging:
            loggers:
              default:
                level: critical
        ''')

        if cls.__configuration__:
            settings.merge(cls.__configuration__)

        # Overriding the db uri becase this is a test session, so db.test_uri
        # will be used instead of the db.uri
        settings.db.url = settings.db.test_url

    @classmethod
    def create_session(cls, *a, expire_on_commit=False, **kw):
        new_session = session_factory(
            bind=cls._engine,
            *a,
            expire_on_commit=expire_on_commit,
            **kw
        )
        cls._sessions.append(new_session)
        return new_session

    @classmethod
    def initialize_orm(cls):
        # Drop the previosely created db if exists.
        with DBManager() as m:
            m.drop_database()
            m.create_database()

        # An engine to create db schema and bind future created sessions
        cls._engine = create_engine()

        # A session factory to create and store session
        # to close it on tear down
        session = cls.create_session(expire_on_commit=True)

        # Creating database objects
        setup_schema(session)
        session.commit()

        # Closing the session to free the connection for future sessions.
        session.close()

        cls.__application__.initialize_orm(cls._engine)

    @classmethod
    def mockup(cls):
        """This is a template method so this is optional to override and you
        haven't call the super when overriding it, because there isn't any.
        """
        pass

    @classmethod
    def cleanup_orm(cls):
        # Closing all sessions created by the test writer
        while True:
            try:
                s = cls._sessions.pop()
                s.close()
            except IndexError:
                break

        DBSession.remove()
        if cls._engine is not None:
            cls._engine.dispose()

        # Dropping the previousely created database
        with DBManager() as m:
            m.drop_database()

    @classmethod
    def setup_class(cls):
        if cls.__application__ is None:
            parameters = {}
            if cls.__controller_factory__ is not None:
                parameters['root'] = cls.__controller_factory__()

            cls.__application__ = cls.__application_factory__(
                'Restfulpy testing application',
                **parameters,
            )

        cls.configure_application()
        try:
            cls.initialize_orm()
            cls.mockup()
        except:
            cls.teardown_class()
            raise

    @classmethod
    def teardown_class(cls):
        cls.cleanup_orm()

    @classmethod
    def _ensure_directory(cls, d):
        if not path.exists(d):
            os.makedirs(d, exist_ok=True)

    @classmethod
    def _get_story_filename(cls, story):
        cls._ensure_directory(cls.__story_directory__)
        title = story.title.lower().replace(' ', '-')
        entity = story.base_call.url.split('/')[2]
        filename = path.join(
            cls.__story_directory__,
            f'{story.base_call.verb}-{entity}--{title}.yml'
        )
        return filename

    @classmethod
    def _get_markdown_filename(cls, story):
        cls._ensure_directory(cls.__api_documentation_directory__)
        title = story.title.lower().replace(' ', '-')
        entity = story.base_call.url.split('/')[2]
        filename = path.join(
            cls.__api_documentation_directory__,
            f'{story.base_call.verb}-{entity}--{title}.md'
        )
        return filename

    def given(self, *a, autodoc=True, **kw):
        if self._authentication_token is not None:
            kw.setdefault('authorization', self._authentication_token)

        if self.__story_directory__:
            kw['autodump'] = self._get_story_filename

        if autodoc and self.__api_documentation_directory__:
            kw['autodoc'] = self._get_markdown_filename

        return Given(self.__application__, *a, **kw)

    def when(self, *a, **kw):
        if self._authentication_token is not None:
            kw.setdefault('authorization', self._authentication_token)
        return when(*a, **kw)

    def login(self, form, url='/apiv1/sessions', verb='POST'):
        with self.given(
                None,
                url,
                verb,
                form=form
            ) as story:
            response = story.response
            assert response.status == '200 OK'
            assert 'token' in response.json
            self._authentication_token = response.json['token']

    def logout(self):
        self._authentication_token = None




import re
from datetime import datetime

import itsdangerous
import redis
import ujson
import user_agents
from nanohttp import context, HTTPBadRequest, settings, HTTPUnauthorized

from restfulpy.principal import JwtPrincipal, JwtRefreshToken


class Authenticator:
    """
    An extendable stateless abstract class for encapsulating all stuff about
    authentication
                                                            +
                                                            |
                                              Yes  +--------+-------+  No
                                              +----+  Token Exists? +-----+
                                              |    +----------------+     |
                                              |                           |
                                              |                           |
                                              |                           |
                                 No  +--------+-------+  Yes              |
                                +----+ Verify Tokens? +------+            |
                                |    +----------------+      |            |
                                |                            |    Continue|
        +---+         Yes +-----+-----+                      |       as   |
        |400+^------------+  Damaged? |                      |     isitor |
        +---+             +-----------+                      |            |
                              No|                            |            |
                                |                            |            |
                           +----+----+                       |            |
                           | Expired |                       |            |
                           +----+----+                       |            |
                                |                            |            |
                           +----+----+                       |            |
        +---+         No   | Refresh |                       |            |
        |401+^-------------+ Token   |                 As Use|            |
        +---+              | Exists? |                       |            |
                           +----+----+                       |            |
                                |                            |            |
                             Yes|                            |            |
                                |                            |            |
        +---+       No +--------+---------+                  |            |
        |400+^---------+Is connection SSL?|                  |            |
        +---+          +--------+---------+                  |            |
                                |                            |            |
                            Yes |                            |            |
                                |                            |            |
                       +--------+---------+                  |            |
                       |                  |                  |            |
                       |                  |                  |            |
                       |   Renew Token    |                  |            |
                       |        &         |                  |            |
                       |   Add to Header  |                  |            |
                       |                  |                  |            |
                       |                  |                  |            |
                       +--------+---------+                  |            |
                                |                            |            |
                                |                            |            |
                                |                            |            |
                                |                            |            |
                                |                            |            |
                                |          +----------+      |            |
                                +----------> WSGI App <------+------------+
                                           +----------+



    """

    token_key = 'HTTP_AUTHORIZATION'
    refresh_token_key = 'refresh-token'
    token_response_header = 'X-New-JWT-Token'
    identity_response_header = 'X-Identity'

    def create_principal(self, member_id=None, session_id=None, **kwargs):  # pragma: no cover
        raise NotImplementedError()

    def create_refresh_principal(self, member_id=None):  # pragma: no cover
        raise NotImplementedError()

    def validate_credentials(self, credentials):  # pragma: no cover
        raise NotImplementedError()

    def setup_response_headers(self, new_principal):
        if self.token_response_header in context.response_headers:
            del context.response_headers[self.token_response_header]
        context.response_headers.add_header(
            self.token_response_header,
            new_principal.dump().decode()
        )

    def try_refresh_token(self, session_id):
        morsel = context.cookies.get(self.refresh_token_key)
        if not morsel:
            return self.bad()

        if settings.jwt.refresh_token.secure \
                and context.request_scheme != 'https':
            raise HTTPBadRequest('not allowed')

        if morsel.value is None or not morsel.value.strip():
            return self.bad()

        refresh_token_encoded = morsel.value
        # Decoding the refresh token
        try:
            refresh_principal = JwtRefreshToken.load(refresh_token_encoded)
            self.ok(
                self.create_principal(
                    member_id=refresh_principal.id,
                    session_id=session_id
                ),
                setup_header=True
            )
        except itsdangerous.SignatureExpired:
            self.bad()
        except itsdangerous.BadData:
            self.bad()
            raise HTTPBadRequest()

    def setup_identity_response_header(self, principal):
        if self.identity_response_header in context.response_headers:
            del context.response_headers[self.identity_response_header]

        if principal is not None:
            context.response_headers.add_header(
                self.identity_response_header,
                str(principal.id)
            )

        if principal is None and self.refresh_token_key in context.cookies:
            del context.cookies[self.refresh_token_key]

    def bad(self):
        context.identity = None
        self.setup_identity_response_header(None)

    def ok(self, principal, setup_header=False):
        context.identity = principal
        self.setup_identity_response_header(principal)
        if setup_header:
            self.setup_response_headers(principal)

    def verify_token(self, encoded_token):
        return JwtPrincipal.load(encoded_token)

    def authenticate_request(self):
        if self.token_key not in context.environ:
            self.bad()
            return

        encoded_token = context.environ[self.token_key]
        if encoded_token is None or not encoded_token.strip():
            self.bad()
            return

        try:
            self.ok(self.verify_token(encoded_token))

        except itsdangerous.SignatureExpired as ex:
            # The token has expired. So we're trying to restore it using
            # refresh-token.
            session_id = ex.payload.get('sessionId')
            if session_id:
                self.try_refresh_token(session_id)
            else:
                self.bad()
                raise HTTPUnauthorized()
        except itsdangerous.BadData:
            # The token is Malformed
            self.bad()
            raise HTTPBadRequest()

    def login(self, credentials):
        member = self.validate_credentials(credentials)
        if member is None:
            return None

        principal = self.create_principal(member.id)

        self.ok(principal)

        context.cookies[self.refresh_token_key] = \
            self.create_refresh_principal(member.id).dump().decode()
        context.cookies[self.refresh_token_key]['max-age'] = \
            settings.jwt.refresh_token.max_age
        context.cookies[self.refresh_token_key]['httponly'] = \
            settings.jwt.refresh_token.httponly
        context.cookies[self.refresh_token_key]['secure'] = \
            settings.jwt.refresh_token.secure
        if 'path' in settings.jwt.refresh_token:
            context.cookies[self.refresh_token_key]['path'] = \
                settings.jwt.refresh_token.path

        return principal

    def logout(self):
        self.bad()


class StatefulAuthenticator(Authenticator):
    """

    Redis data-model:

        sessions: HashMap { session_id: member_id }
        member_id: Set { session_id }
        agents: HashMap {
          session_id: {
            remoteAddress: 127.0.0.1,
            machine: pc,
            os: android 4.2.2,
            agent: chrome 55,
            client: RestfulpyClient-js 3.4.5-alpha14,
            app: Mobile Token (shark) 1.5.4,
            lastActivity: 2015-10-26T07:46:36.615661
          }
        }


    User-Agent structure:

        User-Agent can contains customized token and comment in order to
        describe client and app.

        User-Agent: TOKEN (COMMENT)

        TOKEN: RestfulpyClient-TYPE/VERSION
            exp: RestfulpyClient-js/1.2.3

        COMMENT: (APP-TITLE; APP-NAME; APP-VERSION; APP-LOCALIZATION
                 [; OPTIONAL-OTHER-COMMENTS]*)
            exp: (Mobile Token; shark; 1.2.3-preview2; en-US)

        Tips:
            1. TOKEN must not contain any of CLRs or separators
            2. COMMENTS: any TEXT excluding "(" and ")" and ";"
            3. If you don't use one of standard clients of restfulpy, you can
               use this format as TOKEN: RestfulpyClient-custom/0.0.0

    """

    _redis = None
    sessions_key = 'auth:sessions'
    members_key = 'auth:member:%s'
    session_info_key = 'auth:sessions:%s:info'
    remote_address_key = 'REMOTE_ADDR'
    agent_key = 'HTTP_USER_AGENT'

    @staticmethod
    def create_blocking_redis_client():
        return redis.StrictRedis(
            host=settings.authentication.redis.host,
            port=settings.authentication.redis.port,
            db=settings.authentication.redis.db,
            password=settings.authentication.redis.password
        )

    @property
    def redis(self):
        if self.__class__._redis is None:
            self.__class__._redis = self.create_blocking_redis_client()
        return self.__class__._redis

    @classmethod
    def get_member_sessions_key(cls, member_id):
        return cls.members_key % member_id

    @classmethod
    def get_session_info_key(cls, session_id):
        return cls.session_info_key % session_id

    def verify_token(self, encoded_token):
        principal = super().verify_token(encoded_token)
        if not self.validate_session(principal.session_id):
            raise itsdangerous.SignatureExpired(
                'The token has already invalidated',
                principal.payload
            )
        return principal

    def login(self, credentials):
        principal = super().login(credentials)
        if principal is not None:
            self.register_session(principal.id, principal.session_id)
        return principal

    def logout(self):
        self.unregister_session(context.identity.session_id)
        super().logout()

    def extract_agent_info(self):
        remote_address = machine = os = agent = client = app = None

        if self.remote_address_key in context.environ \
                and context.environ[self.remote_address_key]:
            remote_address = context.environ[self.remote_address_key]

        if self.agent_key in context.environ:
            agent_string = context.environ[self.agent_key]
            user_agent = user_agents.parse(agent_string)

            machine = user_agent.is_pc and 'PC' or user_agent.device.family
            os = ' '.join([
                user_agent.os.family,
                user_agent.os.version_string
            ]).strip()
            agent = ' '.join([
                user_agent.browser.family,
                user_agent.browser.version_string
            ]).strip()

            matched_client = re.match(
                '.*RestfulpyClient-(?P<type>.+)/(?P<version>.+) '
                '\((?P<features>.+)\).*',
                agent_string
            )
            if matched_client:
                matched_client = matched_client.groupdict()
                client_type = matched_client['type']
                client_version = matched_client['version']
                # exp: "RestfulpyClient-js 1.2.3-rc10"
                client = f'RestfulpyClient-{client_type} {client_version}'

                features = matched_client['features'].split(';')
                if len(features) >= 3:
                    # exp: "MobileToken (shark) 1.2.3"
                    app = \
                        f'{features[0].strip()} ({features[1].strip()}) ' \
                        f'{features[2].strip()}'

        return {
            'remoteAddress': remote_address or 'NA',
            'machine': machine or 'Other',
            'os': os or 'Other',
            'agent': agent or 'Other',
            'client': client or 'Unknown',
            'app': app or 'Unknown',
            'lastActivity': datetime.utcnow().isoformat()
        }

    def register_session(self, member_id, session_id):
        self.redis.hset(self.sessions_key, session_id, member_id)
        self.redis.sadd(self.get_member_sessions_key(member_id), session_id)
        self.redis.set(
            self.get_session_info_key(session_id),
            ujson.dumps(self.extract_agent_info())
        )

    def unregister_session(self, session_id=None):
        session_id = session_id or context.identity.session_id
        member_id = self.redis.hget(self.sessions_key, session_id)
        self.redis.srem(self.get_member_sessions_key(member_id), session_id)
        self.redis.hdel(self.sessions_key, session_id)
        self.redis.delete(self.get_session_info_key(session_id))

    def invalidate_member(self, member_id=None):
        # store current session id if available
        current_session_id = \
            None if context.identity is None else context.identity.session_id
        while True:
            session_id = self.redis.spop(
                self.get_member_sessions_key(member_id)
            )
            if not session_id:
                break
            self.redis.hdel(self.sessions_key, session_id)
            self.redis.delete(self.get_session_info_key(session_id))
        self.redis.delete(self.get_member_sessions_key(member_id))
        if current_session_id:
            self.try_refresh_token(current_session_id)

    def validate_session(self, session_id):
        return self.redis.hexists(self.sessions_key, session_id)

    def get_member_id_by_session(self, session_id):
        return int(self.redis.hget(self.sessions_key, session_id))

    def ok(self, principal, setup_header=False):
        super().ok(principal, setup_header)
        self.update_session_info(principal.session_id)

    def update_session_info(self, session_id):
        self.redis.set(
            self.get_session_info_key(session_id),
            ujson.dumps(self.extract_agent_info())
        )

    def get_session_info(self, session_id):
        info = self.redis.get(self.get_session_info_key(session_id))
        if info:
            return ujson.loads(info)
        return None




from datetime import date, time

import pytest
from nanohttp import HTTPBadRequest, settings, HTTPStatus, RequestValidator
from nanohttp.contexts import Context
from sqlalchemy import Integer, Unicode, ForeignKey, Boolean, Table, Date,\
    Time, Float
from sqlalchemy.orm import synonym

from restfulpy.orm import DeclarativeBase, Field, relationship, composite, \
    ModifiedMixin


class Actor(DeclarativeBase):
    __tablename__ = 'actor'

    id = Field(
        Integer,
        primary_key=True,
        readonly='709 id is readonly'
    )
    email = Field(
        Unicode(100),
        not_none='701 email cannot be null',
        required='702 email required',
        pattern=(
            '(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)',
            '707 Invalid email address'
        ),
        watermark='Email',
        example="user@example.com"
    )
    title = Field(
        Unicode(50),
        index=True,
        min_length=(4, '703 title must be at least 4 characters'),
        max_length=(50, '704 Maximum allowed length for title is 50'),
        watermark='First Name',
        nullable=True,
    )
    age = Field(
        Integer,
        nullable=True,
        python_type=(int, '705 age must be integer'),
        minimum=(18, '706 age must be greater than 17'),
        maximum=(99, '706 age must be less than 100')
    )



def test_validation(db):
    session = db()
    email = 'user@example.com'

    validate = RequestValidator(Actor.create_validation_rules(strict=True))
    values, querystring = validate(dict(
        email=email,
    ))
    assert values['email'] == email

    # Not None
    with pytest.raises(HTTPStatus) as ctx:
        validate(dict(email=None))
    assert issubclass(ctx.type, HTTPStatus)
    assert isinstance(ctx.value, HTTPStatus)
    assert str(ctx.value) == '701 email cannot be null'

    # Required
    with pytest.raises(HTTPStatus) as ctx:
        validate(dict(title='required-test-case'))
    assert issubclass(ctx.type, HTTPStatus)
    assert isinstance(ctx.value, HTTPStatus)
    assert str(ctx.value) == '702 email required'

    # Minimum length
    with pytest.raises(HTTPStatus) as ctx:
        validate(dict(email=email, title='abc'))
    assert issubclass(ctx.type, HTTPStatus)
    assert isinstance(ctx.value, HTTPStatus)
    assert str(ctx.value) == '703 title must be at least 4 characters'

    # Mamimum length
    with pytest.raises(HTTPStatus) as ctx:
        validate(dict(email=email, title='a'*(50+1)))
    assert issubclass(ctx.type, HTTPStatus)
    assert isinstance(ctx.value, HTTPStatus)
    assert str(ctx.value) == '704 Maximum allowed length for title is 50'

    # Type
    with pytest.raises(HTTPStatus) as ctx:
        validate(dict(email=email, title='abcd', age='a'))
    assert issubclass(ctx.type, HTTPStatus)
    assert isinstance(ctx.value, HTTPStatus)
    assert str(ctx.value) == '705 age must be integer'

    # Minimum
    with pytest.raises(HTTPStatus) as ctx:
        validate(dict(email=email, title='abcd', age=18-1))
    assert issubclass(ctx.type, HTTPStatus)
    assert isinstance(ctx.value, HTTPStatus)
    assert str(ctx.value) == '706 age must be greater than 17'

    # Maximum
    with pytest.raises(HTTPStatus) as ctx:
        validate(dict(email=email, title='abcd', age=99+1))
    assert issubclass(ctx.type, HTTPStatus)
    assert isinstance(ctx.value, HTTPStatus)
    assert str(ctx.value) == '706 age must be less than 100'

    # Pattern
    with pytest.raises(HTTPStatus) as ctx:
        validate(dict(email='invalidemail'))
    assert issubclass(ctx.type, HTTPStatus)
    assert isinstance(ctx.value, HTTPStatus)
    assert str(ctx.value) == '707 Invalid email address'

    # Readonly
    with pytest.raises(HTTPStatus) as ctx:
        validate(dict(email=email, id=22))
    assert issubclass(ctx.type, HTTPStatus)
    assert isinstance(ctx.value, HTTPStatus)
    assert str(ctx.value) == '709 id is readonly'




import time

from bddrest.authoring import response

from restfulpy.mockup import MockupApplication
from restfulpy.authentication import StatefulAuthenticator
from restfulpy.principal import JwtPrincipal, JwtRefreshToken
from restfulpy.testing import ApplicableTestCase


roles = ['admin', 'test']


class MockupAuthenticator(StatefulAuthenticator):
    def validate_credentials(self, credentials):
        raise NotImplementedError()

    def create_refresh_principal(self, member_id=None):
        return JwtRefreshToken(dict(
            id=member_id
        ))

    def create_principal(self, member_id=None, session_id=None, **kwargs):
        return JwtPrincipal(
            dict(id=1, email='test@example.com', roles=roles, sessionId='1')
        )


class TestRefreshTokenWithoutSSl(ApplicableTestCase):
    __application__ = MockupApplication(
        'MockupApplication',
        None,
        authenticator=MockupAuthenticator()
    )

    __configuration__ = ('''
        jwt:
          max_age: .3
          refresh_token:
            max_age: 3
            secure: true
    ''')

    def test_refresh_token_security(self):
        principal = self.__application__.__authenticator__.create_principal()

        token = principal.dump().decode("utf-8")
        refresh_principal = self.__application__.__authenticator__.\
            create_refresh_principal()
        refresh_token = 'refresh-token=' + refresh_principal.dump().\
            decode("utf-8")
        assert refresh_token.startswith('refresh-token=') is True
        self._authentication_token = token

        time.sleep(1)

        with self.given(
            'Refresh tokn can not be set in not secure connections',
            headers={'Cookie': refresh_token},
        ):
            assert response.status == 400




from os import path

from nanohttp import configure as nanohttp_configure, settings


__builtin_config = """

debug: true

# Default timezone.
# empty for local time
# 0, utc, UTC, z or Z for the UTC,
# UTC±HH:MM for specify timezone. ie. +3:30 for tehran
# An instance of datetime.tzinfo is also acceptable
# Example:
# timezone: !!python/object/apply:datetime.timezone
#   - !!python/object/apply:datetime.timedelta [0, 7200, 0]
#   - myzone
timezone:

db:
  # The main uri
  url: sqlite:///devdata.db
  # url: postgresql://postgres:postgres@localhost/restfulpy_demo_dev

  # Will be used to create and drop database(s).
  # administrative_url: postgresql://postgres:postgres@localhost/postgres
  # test_url: postgresql://postgres:postgres@localhost/restfulpy_test
  echo: false

migration:
  directory: migration
  ini: alembic.ini

jwt:
  secret: JWT-SECRET
  algorithm: HS256
  max_age: 86400  # 24 Hours
  refresh_token:
    secret: JWT-REFRESH-SECRET
    algorithm: HS256
    max_age: 2678400  # 30 Days
    secure: true
    httponly: false
    # path: optional
    #path: /

messaging:
  # default_messenger: restfulpy.messaging.providers.SmtpProvider
  default_messenger: restfulpy.messaging.ConsoleMessenger
  default_sender: restfulpy
  mako_modules_directory:
  template_dirs:
    - %(restfulpy_root)s/messaging/templates

templates:
  directories: []

authentication:
  redis:
    host: localhost
    port: 6379
    password: ~
    db: 0

worker:
  gap: .5
  number_of_threads: 1

smtp:
  host: smtp.example.com
  port: 587
  username: user@example.com
  password: password
  local_hostname: localhost
  tls: true
  auth: true
  ssl: false

# Logging stuff
logging:
  loggers:

    default:
      handlers:
        - console
      level: debug
      formatter: default
      propagate: true

    root:
      level: debug
      formatter: default

  handlers:

    default:
      level: notset
      max_bytes: 52428800
      formatter: default

    console:
      type: console

  formatters:
    default:
      format: "%%(asctime)s - %%(name)s - %%(levelname)s - %%(message)s"
      date_format: "%%Y-%%m-%%d %%H:%%M:%%S"

"""


def configure(config=None, files=None, context=None, force=False):

    context = context or {}
    context['restfulpy_root'] = path.dirname(__file__)

    nanohttp_configure(
        context=context,
        force=force
    )
    settings.merge(__builtin_config)

    if config:
        settings.merge(config)

    if files:
        for f in files:
            settings.load_files(f)




import base64
import io
import os
from os.path import split

from .mimetypes_ import guess_type


def encode_multipart_data(fields: dict = None, files: dict = None):  # pragma: no cover
    boundary = ''.join([
        '-----',
        base64.urlsafe_b64encode(os.urandom(27)).decode()
    ])
    crlf = b'\r\n'
    lines = []

    if fields:
        for key, value in fields.items():
            lines.append('--' + boundary)
            lines.append('Content-Disposition: form-data; name="%s"' % key)
            lines.append('')
            lines.append(value)

    if files:
        for key, file_path in files.items():
            filename = split(file_path)[1]
            lines.append('--' + boundary)
            lines.append(
                'Content-Disposition: form-data; name="%s"; filename="%s"' %
                (key, filename))
            lines.append(
                'Content-Type: %s' %
                (guess_type(filename) or 'application/octet-stream'))
            lines.append('')
            lines.append(open(file_path, 'rb').read())

    lines.append('--' + boundary + '--')
    lines.append('')

    body = io.BytesIO()
    length = 0
    for l in lines:
        line = (l if isinstance(l, bytes) else l.encode()) + crlf
        length += len(line)
        body.write(line)
    body.seek(0)
    content_type = 'multipart/form-data; boundary=%s' % boundary
    return content_type, body, length






from nanohttp import settings
from sqlalchemy import Integer, ForeignKey, Unicode
from sqlalchemy.ext.declarative import declared_attr

from ..logging_ import get_logger
from ..orm import Field, FakeJson,synonym
from ..taskqueue import RestfulpyTask
from .providers import create_messenger


logger = get_logger('messaging')


class Email(RestfulpyTask):
    __tablename__ = 'email'


    template_filename = Field(Unicode(200), nullable=True)
    to = Field(Unicode(100), json='to')
    subject = Field(Unicode(256), json='subject')
    cc = Field(Unicode(100), nullable=True, json='cc')
    bcc = Field(Unicode(100), nullable=True, json='bcc')
    _body = Field('body', FakeJson)

    from_ = Field(
        Unicode(100),
        json='from',
        default=lambda: settings.messaging.default_sender
    )

    __mapper_args__ = {
        'polymorphic_identity': __tablename__
    }

    def _set_body(self, body):
        self._body = body

    def _get_body(self):
        return self._body

    @declared_attr
    def body(cls):
        return synonym(
            '_body',
            descriptor=property(cls._get_body, cls._set_body)
        )

    @declared_attr
    def id(cls):
        return Field(
            Integer,
            ForeignKey('restfulpy_task.id'),
            primary_key=True, json='id'
        )

    def do_(self, attachments=None):
        messenger = create_messenger()
        messenger.send(
            self.to,
            self.subject,
            self.body,
            cc=self.cc,
            bcc=self.bcc,
            template_filename=self.template_filename,
            from_=self.from_,
            attachments=attachments
        )

        logger.info('%s is sent to %s', self.subject, self.to)




import sys

from restfulpy.cli import RequireSubCommand, Launcher


class ConfigurationLauncher(Launcher, RequireSubCommand):

    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(
            'configuration',
            help='Configuration tools'
        )
        subparsers = parser.add_subparsers(
            title='configuration command',
            dest='config_command'
        )
        EncryptConfigurationLauncher.register(subparsers)
        DecryptConfigurationLauncher.register(subparsers)
        return parser


class EncryptConfigurationLauncher(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('encrypt', help='Encrypt configu file')
        return parser

    def launch(self):
        sys.stdout.buffer.write(b'#enc')
        sys.stdout.buffer.write(
        self.args.application.__configuration_cipher__.encrypt(
            sys.stdin.buffer.read()
        ))


class DecryptConfigurationLauncher(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(
            'decrypt',
            help='Decrypt the config file'
        )
        return parser

    def launch(self):
        sys.stdout.buffer.write(
            self.args.application.__configuration_cipher__.decrypt(
                sys.stdin.buffer.read().lstrip(b'#enc')
            )
        )




from bddrest import response, when
from nanohttp import json, RestController, context
from sqlalchemy import Unicode, Integer

from restfulpy.controllers import JsonPatchControllerMixin
from restfulpy.orm import commit, DeclarativeBase, Field, DBSession
from restfulpy.testing import ApplicableTestCase


class CommitCheckingModel(DeclarativeBase):
    __tablename__ = 'commit_checking_model'
    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50), unique=True)


class Root(JsonPatchControllerMixin, RestController):

    @json
    @commit
    def post(self):
        m = CommitCheckingModel()
        m.title = context.form['title']
        DBSession.add(m)
        return m

    @json
    def get(self, title: str=None):
        m = DBSession.query(CommitCheckingModel).\
            filter(CommitCheckingModel.title == title).one()
        return m

    @json
    @commit
    def error(self):
        m = CommitCheckingModel()
        m.title = 'Error'
        DBSession.add(m)
        raise Exception()


class TestCommitDecorator(ApplicableTestCase):
    __controller_factory__ = Root

    def test_commit_decorator(self):
        with self.given(
            'Testing the operation of commit decorator',
            verb='POST',
            url='/',
            form=dict(title='first')
        ):
            when('Geting the result of appling commit decorator',
                 verb='GET',
                 url='/first'
                 )
            assert response.json['title'] == 'first'
            assert response.json['id'] == 1

    def test_commit_decorator_and_json_patch(self):
        with self.given(
            'The commit decorator should not to do anything if the request\
            is a jsonpatch.',
            verb='PATCH',
            url='/',
            json=[
                dict(op='post', path='', value=dict(title='second')),
                dict(op='post', path='', value=dict(title='third'))
            ]):
            when('Inset form parameter to body', verb='GET', url='/third')
            assert response.json['title'] == 'third'

    def test_rollback(self):
        with self.given('Raise exception', verb='ERROR', url='/'):
            assert response.status == 500




import pytest

from restfulpy.testing import TestCase


class TestCaseFoo(TestCase):
    _setup_call_count = 0
    _teardown_call_count = 0

    @classmethod
    def setup_class(cls):
        cls._setup_call_count += 1

    @classmethod
    def teardown_class(cls):
        cls._teardown_call_count += 1

    def test_setup_teardown_first(self):
        assert self._setup_call_count == 1
        assert self._teardown_call_count == 0

    @pytest.mark.last
    def test_setup_teardown_second(self):
        """The reason of this test case is to ensure the setup and teardown
        methods was called exactly once
        """
        assert self._setup_call_count == 1
        assert self._teardown_call_count == 0


def test_last():
    assert TestCaseFoo._teardown_call_count == 1




import pytest
from nanohttp import HTTPBadRequest
from nanohttp.contexts import Context
from sqlalchemy import Integer, Unicode

from restfulpy.orm import DeclarativeBase, Field, FilteringMixin


class FilteringObject(FilteringMixin, DeclarativeBase):
    __tablename__ = 'filtering_object'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))


def test_filtering_mixin(db):
    session = db()

    for i in range(1, 6):
        session.add(FilteringObject(
            title='object %s' % i,
        ))

    session.add(FilteringObject(
        title='A simple title',
    ))
    session.commit()

    query = session.query(FilteringObject)

    with Context({'QUERY_STRING': 'id=1'}) as context, \
            pytest.raises(HTTPBadRequest):
        context.query['id'] = 1
        FilteringObject.filter_by_request(query)

    # IN
    with Context({'QUERY_STRING': 'id=IN(1,2,3)'}):
        assert FilteringObject.filter_by_request(query).count() == 3

    # NOT IN
    with Context({'QUERY_STRING': 'id=!IN(1,2,3)'}):
        assert FilteringObject.filter_by_request(query).count() == 3

    # IN (error)
    with Context({'QUERY_STRING': 'id=IN()'}), \
        pytest.raises(HTTPBadRequest):
        FilteringObject.filter_by_request(query)

    # Between
    with Context({'QUERY_STRING': 'id=BETWEEN(1,3)'}):
        assert FilteringObject.filter_by_request(query).count() == 3

    # IS NULL
    with Context({'QUERY_STRING': 'title=null'}):
        assert FilteringObject.filter_by_request(query).count() == 0

    # IS NOT NULL
    with Context({'QUERY_STRING': 'title=!null'}):
        assert FilteringObject.filter_by_request(query).count() == 6

    # ==
    with Context({'QUERY_STRING': 'id=1'}):
        assert FilteringObject.filter_by_request(query).count() == 1

    # !=
    with Context({'QUERY_STRING': 'id=!1'}):
        assert FilteringObject.filter_by_request(query).count() == 5

    # >=
    with Context({'QUERY_STRING': 'id=>=2'}):
        assert FilteringObject.filter_by_request(query).count() == 5

    # >
    with Context({'QUERY_STRING': 'id=>2'}):
        assert FilteringObject.filter_by_request(query).count() == 4

    # <=
    with Context({'QUERY_STRING': 'id=<=3'}):
        FilteringObject.filter_by_request(query).count() == 3

    # <
    with Context({'QUERY_STRING': 'id=<3'}):
        assert FilteringObject.filter_by_request(query).count() == 2

    # LIKE
    with Context({'QUERY_STRING': 'title=%obj%'}):
        assert FilteringObject.filter_by_request(query).count() == 5

    with Context({'QUERY_STRING': 'title=%OBJ%'}):
        assert FilteringObject.filter_by_request(query).count() == 0

    # ILIKE
    with Context({'QUERY_STRING': 'title=~%obj%'}):
        assert FilteringObject.filter_by_request(query).count() == 5

    with Context({'QUERY_STRING': 'title=~%OBJ%'}):
        assert FilteringObject.filter_by_request(query).count() == 5

    with Context({'QUERY_STRING': 'title=A sim%'}):
        assert FilteringObject.filter_by_request(query).count() == 1

    with Context({'QUERY_STRING': 'title=%25ect 5'}):
        assert FilteringObject.filter_by_request(query).count() == 1

    with Context({'QUERY_STRING': 'title=%imple%'}):
        assert FilteringObject.filter_by_request(query).count() == 1

    with Context({'QUERY_STRING': 'title=~%IMPLE%'}):
        assert FilteringObject.filter_by_request(query).count() == 1



import threading

from restfulpy.taskqueue import RestfulpyTask, worker


awesome_task_done = threading.Event()
another_task_done = threading.Event()


class AwesomeTask(RestfulpyTask):

    __mapper_args__ = {
        'polymorphic_identity': 'awesome_task'
    }

    def do_(self, context):
        awesome_task_done.set()


class AnotherTask(RestfulpyTask):

    __mapper_args__ = {
        'polymorphic_identity': 'another_task'
    }

    def do_(self, context):
        another_task_done.set()


class BadTask(RestfulpyTask):

    __mapper_args__ = {
        'polymorphic_identity': 'bad_task'
    }

    def do_(self, context):
        raise Exception()


def test_worker(db):
    session = db()
    awesome_task = AwesomeTask()
    session.add(awesome_task)

    another_task = AnotherTask()
    session.add(another_task)

    bad_task = BadTask()
    session.add(bad_task)

    session.commit()

    tasks = worker(tries=0, filters=RestfulpyTask.type == 'awesome_task')
    assert len(tasks) == 1

    assert awesome_task_done.is_set() == True
    assert another_task_done.is_set() == False

    session.refresh(awesome_task)
    assert awesome_task.status == 'success'

    tasks = worker(tries=0, filters=RestfulpyTask.type == 'bad_task')
    assert len(tasks) == 1
    bad_task_id = tasks[0][0]
    session.refresh(bad_task)
    assert bad_task.status == 'failed'

    tasks = worker(tries=0, filters=RestfulpyTask.type == 'bad_task')
    assert len(tasks) == 0

    # Reset the status of one task
    session.refresh(bad_task)
    bad_task.status = 'in-progress'
    session.commit()
    session.refresh(bad_task)

    RestfulpyTask.reset_status(bad_task_id, session)
    session.commit()
    tasks = worker(tries=0, filters=RestfulpyTask.type == 'bad_task')
    assert len(tasks) == 1

    tasks = worker(tries=0, filters=RestfulpyTask.type == 'bad_task')
    assert len(tasks) == 0

    # Cleanup all tasks
    RestfulpyTask.cleanup(session, statuses=('in-progress', 'failed'))
    session.commit()

    tasks = worker(tries=0, filters=RestfulpyTask.type == 'bad_task')
    assert len(tasks) == 1

    tasks = worker(tries=0, filters=RestfulpyTask.type == 'bad_task')
    assert len(tasks) == 0

    # Doing all remaining tasks
    tasks = worker(tries=0)
    assert len(tasks) == 1





import time
import traceback
from datetime import datetime

from nanohttp import settings
from sqlalchemy import Integer, Enum, Unicode, DateTime
from sqlalchemy.sql.expression import text

from restfulpy.exceptions import RestfulException
from restfulpy.logging_ import get_logger
from restfulpy.orm import TimestampMixin, DeclarativeBase, Field, DBSession, \
    create_thread_unsafe_session


logger = get_logger('taskqueue')


class TaskPopError(RestfulException):
    pass


class RestfulpyTask(TimestampMixin, DeclarativeBase):
    __tablename__ = 'restfulpy_task'

    id = Field(Integer, primary_key=True, json='id')
    priority = Field(Integer, nullable=False, default=50, json='priority')
    status = Field(
        Enum(
            'new',
            'success',
            'in-progress',
            'failed',
            name='task_status_enum'
        ),
        default='new',
        nullable=True, json='status'
    )
    fail_reason = Field(Unicode(2048), nullable=True, json='reason')
    started_at = Field(DateTime, nullable=True, json='startedAt')
    terminated_at = Field(DateTime, nullable=True, json='terminatedAt')
    type = Field(Unicode(50))

    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
        'polymorphic_on': type
    }

    def do_(self):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def pop(cls, statuses={'new'}, filters=None, session=DBSession):

        find_query = session.query(
            cls.id.label('id'),
            cls.created_at,
            cls.status,
            cls.type,
            cls.priority
        )
        if filters is not None:
            find_query = find_query.filter(
                text(filters) if isinstance(filters, str) else filters
            )

        find_query = find_query \
            .filter(cls.status.in_(statuses)) \
            .order_by(cls.priority.desc()) \
            .order_by(cls.created_at) \
            .limit(1) \
            .with_for_update()

        cte = find_query.cte('find_query')

        update_query = RestfulpyTask.__table__.update() \
            .where(RestfulpyTask.id == cte.c.id) \
            .values(status='in-progress') \
            .returning(RestfulpyTask.__table__.c.id)

        task_id = session.execute(update_query).fetchone()
        session.commit()
        if not task_id:
            raise TaskPopError('There is no task to pop')
        task_id = task_id[0]
        task = session.query(cls).filter(cls.id == task_id).one()
        return task

    def execute(self, context, session=DBSession):
        try:
            isolated_task = session \
                .query(RestfulpyTask) \
                .filter(RestfulpyTask.id == self.id) \
                .one()
            isolated_task.do_(context)
            session.commit()
        except:
            session.rollback()
            raise

    @classmethod
    def cleanup(cls, session=DBSession, statuses=['in-progress']):
        session.query(RestfulpyTask) \
            .filter(RestfulpyTask.status.in_(statuses)) \
            .with_lockmode('update') \
            .update({
                'status': 'new',
                'started_at': None,
                'terminated_at': None
            }, synchronize_session='fetch')

    @classmethod
    def reset_status(cls, task_id, session=DBSession,
                     statuses=['in-progress']):
        session.query(RestfulpyTask) \
            .filter(RestfulpyTask.status.in_(statuses)) \
            .filter(RestfulpyTask.id == task_id) \
            .with_lockmode('update') \
            .update({
                'status': 'new',
                'started_at': None,
                'terminated_at': None
            }, synchronize_session='fetch')


def worker(statuses={'new'}, filters=None, tries=-1):
    isolated_session = create_thread_unsafe_session()
    context = {'counter': 0}
    tasks = []

    while True:
        context['counter'] += 1
        logger.debug("Trying to pop a task, Counter: %s" % context['counter'])
        try:
            task = RestfulpyTask.pop(
                statuses=statuses,
                filters=filters,
                session=isolated_session
            )
            assert task is not None

        except TaskPopError as ex:
            logger.debug('No task to pop: %s' % ex.to_json())
            isolated_session.rollback()
            if tries > -1:
                tries -= 1
                if tries <= 0:
                    return tasks
            time.sleep(settings.worker.gap)
            continue
        except:
            logger.exception('Error when popping task.')
            raise

        try:
            task.execute(context)

            # Task success
            task.status = 'success'
            task.terminated_at = datetime.utcnow()

        except:
            logger.exception('Error when executing task: %s' % task.id)
            task.status = 'failed'
            task.fail_reason = traceback.format_exc()

        finally:
            if isolated_session.is_active:
                isolated_session.commit()
            tasks.append((task.id, task.status))




from nanohttp import configure, settings

from restfulpy.principal import JwtPrincipal


def test_principal():
    __configuration__ = '''
    jwt:
      secret: JWT-SECRET
      algorithm: HS256
      max_age: 86400  # 24 Hours
      refresh_token:
        secret: JWT-REFRESH-SECRET
        algorithm: HS256
        max_age: 2678400  # 30 Days
    '''
    configure(force=True)
    settings.merge(__configuration__)

    principal = JwtPrincipal(dict(
        email='test@example.com',
        id=1,
        sessionId=1,
        roles=['admin']
    ))

    assert principal.email == 'test@example.com'
    assert principal.id == 1
    assert principal.session_id == 1
    assert principal.roles == ['admin']
    assert principal.is_in_roles('admin') is True
    assert principal.is_in_roles('admin', 'god') is True

    encoded = principal.dump()

    principal = JwtPrincipal.load(encoded.decode())
    assert principal.email == 'test@example.com'
    assert principal.id == 1
    assert principal.session_id == 1
    assert principal.roles == ['admin']
    assert principal.is_in_roles('admin') is True
    assert principal.is_in_roles('admin', 'god') is True

    principal = JwtPrincipal.load(encoded.decode(), force=True)
    assert principal.email == 'test@example.com'
    assert principal.id == 1
    assert principal.session_id == 1
    assert principal.roles == ['admin']
    assert principal.is_in_roles('admin') is True
    assert principal.is_in_roles('admin', 'god') is True

    principal =\
        JwtPrincipal.load((b'Bearer %s' % encoded).decode(), force=True)
    assert principal.email == 'test@example.com'
    assert principal.id == 1
    assert principal.session_id == 1
    assert principal.roles == ['admin']
    assert principal.is_in_roles('admin') is True
    assert principal.is_in_roles('admin', 'god') is True




import functools
from datetime import datetime, date, time
from decimal import Decimal

from nanohttp import context, HTTPNotFound, HTTPBadRequest, validate
from sqlalchemy import Column, event
from sqlalchemy.ext.associationproxy import ASSOCIATION_PROXY
from sqlalchemy.ext.hybrid import HYBRID_PROPERTY
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import Query, CompositeProperty, \
    RelationshipProperty
from sqlalchemy.orm.attributes import InstrumentedAttribute

from ..datetimehelpers import parse_datetime, parse_date, parse_time, \
    format_date, format_time, format_datetime
from ..utils import to_camel_case
from .field import Field
from .metadata import MetadataField
from .mixins import PaginationMixin, FilteringMixin, OrderingMixin


class BaseModel(object):

    @classmethod
    def get_column(cls, column):
        if isinstance(column, str):
            mapper = inspect(cls)
            return mapper.columns[column]
        # Commented-out by vahid, because I cannot reach here by tests,
        # I think it's not necessary at all.
        # if isinstance(column, SynonymProperty):
        #     return column.parent.columns[column.name]
        return column

    @classmethod
    def import_value(cls, column, v):
        c = cls.get_column(column)
        if isinstance(c, Column) or isinstance(c, InstrumentedAttribute):
            if c.type.python_type is bool and not isinstance(v, bool):
                return str(v).lower() == 'true'
        return v

    @classmethod
    def get_column_info(cls, column):
        # Use original property for proxies
        if hasattr(column, 'original_property') and column.original_property:
            info = column.info.copy()
            info.update(column.original_property.info)
        else:
            info = column.info

        if not info.get('json'):
            info['json'] = to_camel_case(column.key)

        return info

    @classmethod
    def prepare_for_export(cls, column, v):
        info = cls.get_column_info(column)
        param_name = info.get('json')

        if hasattr(column, 'property') \
                and isinstance(column.property, RelationshipProperty) \
                and column.property.uselist:
            result = [c.to_dict() for c in v]

        elif hasattr(column, 'property') \
            and isinstance(column.property, CompositeProperty):
            result = v.__composite_values__()

        elif v is None:
            result = v

        elif isinstance(v, datetime):
            result = format_datetime(v)

        elif isinstance(v, date):
            result = format_date(v)

        elif isinstance(v, time):
            result = format_time(v)

        elif hasattr(v, 'to_dict'):
            result = v.to_dict()

        elif isinstance(v, Decimal):
            result = str(v)

        else:
            result = v

        return param_name, result

    @classmethod
    def iter_metadata_fields(cls):
        for c in cls.iter_json_columns(
                relationships=True,
                include_readonly_columns=True,
                include_protected_columns=True
            ):
            yield from MetadataField.from_column(
                cls.get_column(c),
                info=cls.get_column_info(c)
            )

    @classmethod
    def json_metadata(cls):
        fields = {f.name: f.to_json() for f in cls.iter_metadata_fields()}
        mapper = inspect(cls)
        return {
            'name': cls.__name__,
            'primaryKeys': [c.key for c in mapper.primary_key],
            'fields': fields
        }

    def update_from_request(self):
        for column, value in self.extract_data_from_request():
            setattr(
                self,
                column.key[1:] if column.key.startswith('_') else column.key,
                self.import_value(column, value)
            )

    @classmethod
    def iter_columns(cls, relationships=True, synonyms=True, composites=True,
                     use_inspection=True, hybrids=True):
        if use_inspection:
            mapper = inspect(cls)
            for k, c in mapper.all_orm_descriptors.items():

                if k == '__mapper__':
                    continue

                if c.extension_type == ASSOCIATION_PROXY:
                    continue

                if (not hybrids and c.extension_type == HYBRID_PROPERTY) \
                        or (not relationships and k in mapper.relationships) \
                        or (not synonyms and k in mapper.synonyms) \
                        or (not composites and k in mapper.composites):
                    continue

                yield getattr(cls, k)

        else:
            for c in cls.__table__.c:
                yield c

    @classmethod
    def iter_json_columns(cls, include_readonly_columns=True,
                          include_protected_columns=False, **kw):
        for c in cls.iter_columns(**kw):

            info = cls.get_column_info(c)
            if (not include_protected_columns and info.get('protected')) or \
                    (not include_readonly_columns and info.get('readonly')):
                continue

            yield c

    @classmethod
    def extract_data_from_request(cls):
        for c in cls.iter_json_columns(
                include_protected_columns=True,
                include_readonly_columns=False
        ):
            info = cls.get_column_info(c)
            param_name = info.get('json')

            if param_name in context.form:

                if hasattr(c, 'property') and hasattr(c.property, 'mapper'):
                    raise HTTPBadRequest('Invalid attribute')

                value = context.form[param_name]

                # Ensuring the python type, and ignoring silently if the
                # python type is not specified
                try:
                    type_ = c.type.python_type
                except NotImplementedError:
                    yield c, value
                    continue

                # Parsing date and or time if required.
                if type_ in (datetime, date, time):
                    try:
                        if type_ == time:
                            yield c, parse_time(value)

                        elif type_ == datetime:
                            yield c, parse_datetime(value)

                        elif type_ == date:
                            yield c, parse_date(value)

                    except ValueError:
                        raise HTTPBadRequest(f'Invalid date or time: {value}')

                else:
                    yield c, value

    def to_dict(self):
        result = {}
        for c in self.iter_json_columns():
            result.setdefault(
                *self.prepare_for_export(c, getattr(self, c.key))
            )
        return result

    @classmethod
    def create_sort_criteria(cls, sort_columns):
        criteria = []
        columns = {
            cls.get_column_info(c).get('json'): c
            for c in cls.iter_json_columns()
        }
        for column_name, option in sort_columns:
            if column_name in columns:
                criteria.append((columns[column_name], option == 'desc'))
        return criteria

    @classmethod
    def filter_paginate_sort_query_by_request(cls, query=None):
        query = query or cls.query

        if issubclass(cls, FilteringMixin):
            query = cls.filter_by_request(query)

        if issubclass(cls, OrderingMixin):
            query = cls.sort_by_request(query)

        if issubclass(cls, PaginationMixin):
            query = cls.paginate_by_request(query=query)

        return query

    @classmethod
    def dump_query(cls, query=None):
        result = []
        for o in cls.filter_paginate_sort_query_by_request(query):
            result.append(o.to_dict())
        return result

    @classmethod
    def expose(cls, func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            if result is None:
                raise HTTPNotFound()
            if isinstance(result, Query):
                return cls.dump_query(result)
            return result

        return wrapper

    @classmethod
    def create_validation_rules(cls, strict=False):
        fields = {}
        for f in cls.iter_metadata_fields():
            fields[f.name] = field = dict(
                required=f.required,
                type_=f.type_,
                minimum=f.minimum,
                maximum=f.maximum,
                pattern=f.pattern,
                min_length=f.min_length,
                max_length=f.max_length,
                not_none=f.not_none,
                readonly=f.readonly
            )

            if not strict and 'required' in field:
                del field['required']
        return fields

    @classmethod
    def validate(cls, strict=False, fields=None):
        if callable(strict):
            # Decorator is used without any parameter and call parentesis.
            func = strict
            strict = False
        else:
            func = None

        rules = cls.create_validation_rules(strict)
        if fields:
            rules.update(fields)

        decorator = validate(**rules)

        if func:
            return decorator(func)

        return decorator




import abc
import os

from Crypto.Cipher import AES


class Cipher(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def encrypt(self, raw):
        pass

    @abc.abstractmethod
    def decrypt(self, enc):
        pass


class AESCipher(Cipher):

    def __init__(self, key, random=os.urandom):
        self.bs = 16
        self.key = key
        self.random = random

    def encrypt(self, raw):
        raw = self._pad(raw)
        iv = self.random(AES.block_size)
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return iv + cipher.encrypt(raw)

    def decrypt(self, enc):
        iv = enc[:AES.block_size]
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        result = self._unpad(cipher.decrypt(enc[AES.block_size:]))
        if not result.strip():
            raise ValueError()
        return result

    def _pad(self, s):
        remaining_bytes = len(s) % self.bs
        padding_bytes = self.bs - remaining_bytes
        return s + padding_bytes * bytes([padding_bytes])

    @staticmethod
    def _unpad(s):
        return s[:-ord(s[len(s)-1:])]





from sqlalchemy import Integer, Unicode

from restfulpy.orm import DeclarativeBase, Field, FullTextSearchMixin, \
    fts_escape, to_tsvector


class FullTextSearchObject(FullTextSearchMixin, DeclarativeBase):
    __tablename__ = 'fulltext_search_object'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))

    __ts_vector__ = to_tsvector(
        title
    )


def test_fts_escape(db):
     result = fts_escape('&%!^$*[](){}\\')
     assert result == '\&\%\!\^\$\*\[\]\(\)\{\}\\\\'


def test_to_tsvector(db):
    result = to_tsvector('a', 'b', 'c')
    assert str(result) == 'to_tsvector(:to_tsvector_1, :to_tsvector_2)'


def test_activation_mixin(db):
    session = db()
    query = session.query(FullTextSearchObject)
    query = FullTextSearchObject.search('a', query)
    query_str = str(query)

    assert str(query) == \
        'SELECT fulltext_search_object.id AS fulltext_search_object_id, '\
        'fulltext_search_object.title AS fulltext_search_object_title \nFROM '\
        'fulltext_search_object \nWHERE to_tsvector(%(to_tsvector_1)s, '\
        'fulltext_search_object.title) @@ to_tsquery(%(to_tsvector_2)s)'




import os
import sys
from os import path

from nanohttp import quickstart

from restfulpy.cli import Launcher
from restfulpy.utils import to_pascal_case
import restfulpy


DEFAULT_ADDRESS = '8080'
HERE = path.dirname(__file__)
TEMPLATES_PATH = path.abspath(path.join(HERE, 'templates'))
TEMPLATES = [
    d for d in os.listdir(TEMPLATES_PATH)
    if path.isdir(path.join(TEMPLATES_PATH, d)) and d[0] not in ('.', '_')
]


class ScaffoldLauncher(Launcher):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(
            'scaffold',
            help='Creates an empty boilerplate'
        )
        parser.add_argument(
            'name',
            help='The snake_case project\'s name'
        )
        parser.add_argument(
            'author',
            help='The project\'s author'
        )
        parser.add_argument(
            'email',
            help='The projects author\'s email'
        )
        parser.add_argument(
            '-t',
            '--template',
            default='full',
            help= \
                f'The project\'s template, one of {", ".join(TEMPLATES)}. '
                f'default: full.'
        )
        parser.add_argument(
            '-o',
            '--overwrite',
            default=False,
            action='store_true',
            help= \
                'Continue and overwrite files when the target '
                'directory(-d/--directory) is not empty.'
        )
        parser.add_argument(
            '-d',
            '--directory',
            default='.',
            help= \
                'Change to this directory before generating new files. It '
                'will make it if does not exists. default: "."'
        )

        return parser

    def launch(self):
        template_path = path.join(TEMPLATES_PATH, self.args.template)
        target_path = path.abspath(self.args.directory)
        title_snakecase = self.args.name.lower()

        if not path.exists(template_path):
            print(f'Invalid template: {template_path}', file=sys.stderr)
            return 1

        if not path.exists(target_path):
            os.makedirs(target_path)

        for top, subdirectories, subfiles in os.walk(template_path):
            current_directory = path.relpath(top, template_path) \
                .replace('__project__', title_snakecase)
            for filename in subfiles:
                if not filename.endswith('.template'):
                    continue

                source = path.abspath(path.join(top, filename))
                target = path.abspath(path.join(
                    target_path,
                    current_directory,
                    filename[:-9].replace('__project__', title_snakecase)
                ))
                if not self.args.overwrite and path.exists(target):
                    print(
                        f'Target file exists: {target}, use -o to overwrite',
                        file=sys.stderr
                    )
                    return 1

                os.makedirs(path.dirname(target), exist_ok=True)
                self.install_file(source, target)
                if self.args.template == 'singlefile':
                    os.chmod(target, 0o774)

    def install_file(self, source, target):
        print(f'Installing  {target}')
        title_snakecase = self.args.name.lower()
        title_camelcase = to_pascal_case(title_snakecase)

        with open(source) as s, open(target, 'w') as t:
            content = s.read()
            content = content.replace('${project_snakecase}', title_snakecase)
            content = content.replace('${project_author}', self.args.author)
            content = content.replace(
                '${project_author_email}',
                self.args.email
            )
            content = content.replace(
                '${project_camelcase}',
                title_camelcase,
            )
            content = content.replace(
                '${restfulpy_version}',
                restfulpy.__version__
            )
            t.write(content)




import sys
from os import makedirs, path
from os.path import join

from nanohttp import settings
from sqlalchemy import MetaData

from restfulpy.cli import Launcher


class EntityRelationshipDiagrams(Launcher):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(
            'entity-relationship-diagram',
            help='Generate erd from database. '
                 'Make sure database and tables exists.'
        )

        parser.add_argument(
            '-u',
            '--url',
            type=str,
            default=None,
            help='DB url.'
        )

        parser.add_argument(
            '-o',
            '--output',
            type=str,
            default='data/class-diagrams/schema.png',
            help='Output file name.'
        )

        parser.add_argument(
            '-d',
            '--datatype',
            type=bool,
            default=True,
            help='Show datatypes.'
        )

        parser.add_argument(
            '-i',
            '--index',
            type=bool,
            default=True,
            help='Show indexes.'
        )

        parser.add_argument(
            '-c',
            '--concentrate',
            type=bool,
            default=True,
            help='Concentrate.'
        )

        parser.add_argument(
            '-r',
            '--rankdir',
            type=str,
            default='LR',
            help='Rankdir.'
        )

        return parser

    def launch(self):
        try:
            from sqlalchemy_schemadisplay import create_schema_graph
        except ImportError:
            print(
                'You have to install sqlalchemy_schemadisplay to use this '
                'feature',
                file=sys.stderr
            )
            sys.exit(1)

        db_url = self.args.url if self.args.url else settings.db.url
        try:
            metadata = MetaData(db_url)
            graph = create_schema_graph(
                metadata=metadata,
                show_datatypes=self.args.datatype,
                show_indexes=self.args.index,
                rankdir=self.args.rankdir,
                concentrate=self.args.concentrate)

        except Exception as ex:
            print(ex)
            return

        output_dir = \
            join(self.args.application.root_path, 'data/class-diagrams/')
        output = join(output_dir, self.args.output)
        makedirs(path.dirname(output), exist_ok=True)

        try:
            graph.write_png(output)
            print('ERD generated in', output)
            print('If it is empty, probably you need to create schema of db.')
        except FileNotFoundError:
            print("You need to install graphviz, see this link: \n"
                  "https://www.graphviz.org/download/")
        except Exception as ex:
            print(ex)




from sqlalchemy import Unicode

from restfulpy.orm import DeclarativeBase, Field, ModifiedMixin


class ModificationCheckingModel(ModifiedMixin, DeclarativeBase):
    __tablename__ = 'modification_checking_model'

    title = Field(Unicode(50), primary_key=True)


def test_modified_mixin(db):
    session = db()

    instance = ModificationCheckingModel(
        title='test title',
    )

    session.add(instance)
    session.commit()

    assert instance.modified_at is None
    assert instance.created_at is not None
    assert instance.last_modification_time == instance.created_at

    instance = session.query(ModificationCheckingModel).one()
    assert instance.modified_at is None
    assert instance.created_at is not None
    assert instance.last_modification_time == instance.created_at

    instance.title = 'Edited title'
    session.commit()
    assert instance.modified_at is not None
    assert instance.created_at is not None
    assert instance.last_modification_time == instance.modified_at

    instance = session.query(ModificationCheckingModel).one()
    assert instance.modified_at is not None
    assert instance.created_at is not None
    assert instance.last_modification_time == instance.modified_at




from datetime import datetime

import pytest
from bddrest import response, when, status
from dateutil.tz import tzoffset
from nanohttp import json
from sqlalchemy import Integer, DateTime

from restfulpy.configuration import settings
from restfulpy.controllers import JsonPatchControllerMixin, ModelRestController
from restfulpy.datetimehelpers import parse_datetime, format_datetime
from restfulpy.mockup import mockup_localtimezone
from restfulpy.orm import commit, DeclarativeBase, Field, DBSession
from restfulpy.testing import ApplicableTestCase


class Metting(DeclarativeBase):
    __tablename__ = 'metting'
    id = Field(Integer, primary_key=True)
    when = Field(DateTime)


class Root(JsonPatchControllerMixin, ModelRestController):
    __model__ = 'metting'

    @json
    @commit
    def post(self):
        m = Metting()
        m.update_from_request()
        DBSession.add(m)
        return m.when.isoformat()


class TestDateTime(ApplicableTestCase):
    __controller_factory__ = Root

    def test_update_from_request(self):
        with self.given(
            'Posting a datetime in form',
            verb='POST',
            form=dict(
                when='2001-01-01T00:01',
            )
        ):
            assert status == 200
            assert response.json == '2001-01-01T00:01:00'

            when(
                'Posting a date instead of datetime',
                form=dict(
                    when='2001-01-01'
                )
            )

            assert status == 200
            assert response.json == '2001-01-01T00:00:00'

            when(
                'Posting an invalid datetime',
                form=dict(
                    when='2001-00-01'
                )
            )

            assert status == '400 Invalid date or time: 2001-00-01'

    def test_naive_datetime_parsing(self):
        # The application is configured to use system's local date and time.
        settings.timezone = None

        # Submit without timezone: accept and assume the local date and time.
        assert datetime(1970, 1, 1) == parse_datetime('1970-01-01T00:00:00')
        assert datetime(1970, 1, 1, microsecond=1000) == \
            parse_datetime('1970-01-01T00:00:00.001')
        assert datetime(1970, 1, 1, microsecond=1000) == \
            parse_datetime('1970-01-01T00:00:00.001000')

        # Timezone aware
        # Submit with 'Z' and or '+3:30':
        # accept and assume as the UTC, so we have to convert
        # it to local date and time before continuing the rest of process
        with mockup_localtimezone(tzoffset(None, 3600)):
            assert datetime(1970, 1, 1, 1) == \
                parse_datetime('1970-01-01T00:00:00Z')
            assert datetime(1970, 1, 1, 1, 30) == \
                parse_datetime('1970-01-01T00:00:00-0:30')

    def test_timezone_aware_datetime_parsing(self):
        # The application is configured to use a specific timezone
        settings.timezone = tzoffset('Tehran', 12600)
        with pytest.raises(ValueError):
            parse_datetime('1970-01-01T00:00:00')

        assert datetime(1970, 1, 1, 3, 30, tzinfo=tzoffset(
            'Tehran', 12600)) == parse_datetime('1970-01-01T00:00:00Z')
        assert datetime(1970, 1, 1, 4, 30, tzinfo=tzoffset(
            'Tehran', 12600)) == parse_datetime('1970-01-01T00:00:00-1:00')

    def test_naive_datetime_formatting(self):
        # The application is configured to use system's local date and time.
        settings.timezone = None

        assert '1970-01-01T00:00:00' == format_datetime(datetime(1970, 1, 1))
        assert '1970-01-01T00:00:00.000001' == \
            format_datetime(datetime(1970, 1, 1, 0, 0, 0, 1))

        with mockup_localtimezone(tzoffset(None, 3600)):
            assert '1970-01-01T00:00:00' == format_datetime(
                datetime(1970, 1, 1, tzinfo=tzoffset(None, 3600))
            )

    def test_timezone_aware_datetime_formatting(self):
        # The application is configured to use a specific timezone as the
        # default
        settings.timezone = tzoffset('Tehran', 12600)

        with pytest.raises(ValueError):
            format_datetime(datetime(1970, 1, 1))

        assert '1970-01-01T00:00:00+03:30' == \
            format_datetime(datetime(1970, 1, 1, tzinfo=tzoffset(None, 12600)))

    def test_naive_unix_timestamp(self):
        # The application is configured to use system's local date and time.
        settings.timezone = None

        assert datetime(1970,1,1,0,0,1,334300) == \
            parse_datetime('1.3343')
        assert datetime(1970,1,1,0,0,1,334300) == \
            parse_datetime('1.3343')

    def test_timezone_aware_unix_timestamp(self):
        # The application is configured to use a specific timezone as the
        # default
        settings.timezone = tzoffset('Tehran', 12600)

        assert datetime(
            1970, 1, 1, 3, 30, 1, 334300, tzinfo=tzoffset('Tehran', 12600)
        ) == parse_datetime('1.3343')




import re

from nanohttp import HTTPBadRequest
from sqlalchemy import Column, Unicode, String
from sqlalchemy.orm import relationship as sa_relationship, \
    composite as sa_composite, synonym as sa_synonym


class Field(Column):

    def __init__(self, *args, json=None, readonly=None, max_length=None,
                 min_length=None, maximum=None, minimum=None, pattern=None,
                 protected=None, watermark=None, not_none=None, nullable=False,
                 required=None, label=None, example=None, default=None,
                 python_type=None,
                 **kwargs):

        info = {
            'json': json,
            'readonly': readonly,
            'protected': protected,
            'watermark': watermark,
            'min_length': min_length,
            'maximum': maximum,
            'minimum': minimum,
            'pattern': pattern,
            'example': example,
            'not_none': not_none,
            'default': default,
            'type': python_type
        }

        if max_length is None and args \
                and isinstance(args[0], (Unicode, String)):
            info['max_length'] = args[0].length
        else:
            info['max_length'] = max_length

        if required is not None:
            info['required'] = required

        if not_none is not None:
            nullable = False

        super(Field, self).__init__(
            *args,
            info=info,
            nullable=nullable,
            default=default,
            **kwargs
        )


def relationship(*args, json=None, protected=True, readonly=True, **kwargs):
    info = {
        'json': json,
        'protected': protected,
        'readonly': readonly,
    }

    return sa_relationship(*args, info=info, **kwargs)


def composite(*args, json=None, protected=None, readonly=None, **kwargs):
    info = {
        'json': json,
        'protected': protected,
        'readonly': readonly,
    }

    return sa_composite(*args, info=info, **kwargs)


def synonym(*args, json=None, protected=None, readonly=None, **kwargs):
    info = {
        'json': json,
        'protected': protected,
        'readonly': readonly,
    }

    return sa_synonym(*args, info=info, **kwargs)




from ..utils import to_camel_case


class FieldInfo:
    def __init__(self, type_=str, pattern=None, max_length=None,
                 min_length=None, readonly=False, protected=False,
                 minimum=None, maximum=None, not_none=None, required=None,
                 default=None):
        self.type_ = type_
        self.pattern = pattern
        self.max_length = max_length
        self.min_length = min_length
        self.minimum = minimum
        self.maximum = maximum
        self.readonly = readonly
        self.protected = protected
        self.not_none = not_none
        self.required = required
        self.default = default

    def to_json(self):
        return {
            'type': self.type_.__name__ if self.type_ else None,
            'not_none': self.not_none,
            'required': self.required,
            'pattern': self.pattern,
            'maxLength': self.max_length,
            'minLength': self.min_length,
            'readonly': self.readonly,
            'protected': self.protected,
            'minimum': self.minimum,
            'maximum': self.maximum,
            'default': self.default,
        }

    def __copy__(self):
        new_one = type(self)()
        new_one.__dict__.update(self.__dict__)
        return new_one

    def to_dict(self):
        return {
            k: v for k, v in self.__dict__.items() if not k.startswith('_')
        }


class MetadataField(FieldInfo):
    def __init__(self, name, key, primary_key=False, label=None,
                 watermark=None, icon=None, message='Invalid value',
                 example=None, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.key = key[1:] if key.startswith('_') else key
        self.primary_key = primary_key
        self.label = label or watermark
        self.watermark = watermark
        self.example = example

    @classmethod
    def from_column(cls, c, info=None):
        if not info:
            info = c.info
        json_name = info.get('json') or to_camel_case(c.key)
        result = []

        key = c.key

        if hasattr(c, 'default') and c.default:
            default = c.default.arg if c.default.is_scalar else None
        else:
            default = None

        result.append(cls(
            json_name,
            key,
            default=default,
            type_=info.get('type'),
            not_none=info.get('not_none'),
            required=info.get('required'),
            pattern=info.get('pattern'),
            max_length=info.get('max_length'),
            min_length=info.get('min_length'),
            minimum=info.get('minimum'),
            maximum=info.get('maximum'),
            watermark=info.get('watermark', None),
            label=info.get('label', None),
            example=info.get('example', None),
            primary_key=hasattr(c.expression, 'primary_key') \
                and c.expression.primary_key,
            readonly=info.get('readonly', False),
            protected=info.get('protected', False)
        ))

        return result

    def to_json(self):
        result = super().to_json()
        result.update(
            name=self.name,
            example=self.example,
            watermark=self.watermark,
            label=self.label,
            key=self.key,
            primaryKey=self.primary_key,
        )
        return result




from bddrest import response, when, Update, status
from nanohttp import json
from sqlalchemy import Unicode, Integer, Date, Float, ForeignKey, Boolean, \
    DateTime
from sqlalchemy.ext.associationproxy import association_proxy

from restfulpy.controllers import JsonPatchControllerMixin, ModelRestController
from restfulpy.orm import commit, DeclarativeBase, Field, DBSession, \
    composite, FilteringMixin, PaginationMixin, OrderingMixin, relationship, \
    ModifiedMixin, ActivationMixin, synonym
from restfulpy.testing import ApplicableTestCase


class FullName(object):  # pragma: no cover
    def __init__(self, first_name, last_name):
        self.first_name = first_name
        self.last_name = last_name

    def __composite_values__(self):
        return '%s %s' % (self.first_name, self.last_name)

    def __repr__(self):
        return "FullName(%s %s)" % (self.first_name, self.last_name)

    def __eq__(self, other):
        return isinstance(other, FullName) and \
               other.first_name == self.first_name and \
               other.last_name == self.last_name

    def __ne__(self, other):
        return not self.__eq__(other)


class Book(DeclarativeBase):
    __tablename__ = 'book'
    id = Field(Integer, primary_key=True)
    title = Field(Unicode(10), nullable=True)
    member_id = Field(Integer, ForeignKey("member.id"))


class Keyword(DeclarativeBase):
    __tablename__ = 'keyword'
    id = Field(Integer, primary_key=True)
    keyword = Field(Unicode(64))


class MemberKeywords(DeclarativeBase):
    __tablename__ = 'member_keywords'
    member_id = Field(Integer, ForeignKey("member.id"), primary_key=True)
    keyword_id = Field(Integer, ForeignKey("keyword.id"), primary_key=True)


class Member(
    ActivationMixin,
    ModifiedMixin,
    FilteringMixin,
    PaginationMixin,
    OrderingMixin,
    DeclarativeBase
):
    __tablename__ = 'member'

    id = Field(Integer, primary_key=True)
    email = Field(
        Unicode(100),
        unique=True,
        index=True,
        json='email',
        pattern=r'(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)',
        watermark='Email',
        example="user@example.com",
    )
    title = Field(
        Unicode(50),
        index=True,
        min_length=2,
        watermark='First Name'
    )
    first_name = Field(
        Unicode(50),
        index=True,
        json='firstName',
        min_length=2,
        watermark='First Name'
    )
    last_name = Field(
        Unicode(100),
        json='lastName',
        min_length=2,
        watermark='Last Name'
    )
    phone = Field(
        Unicode(10), nullable=True, json='phone', min_length=10,
        pattern= \
            r'\d{3}[-\.\s]??\d{3}[-\.\s]??\d{4}|\(\d{3}\)\s*\d{3}[-\.\s]??'
            r'\d{4}|\d{3}[-\.\s]??\d{4}',
        watermark='Phone'
    )
    name = composite(
        FullName,
        first_name,
        last_name,
        readonly=True,
        json='fullName'
    )
    _password = Field(
        'password',
        Unicode(128),
        index=True,
        json='password',
        protected=True,
        min_length=6
    )
    birth = Field(Date)
    weight = Field(Float(asdecimal=True), default=50)
    _keywords = relationship(
        'Keyword',
        secondary='member_keywords',
        protected=False
    )
    keywords = association_proxy(
        '_keywords',
        'keyword',
        creator=lambda k: Keyword(keyword=k)
    )
    visible = Field(Boolean, nullable=True)
    last_login_time = Field(DateTime, json='lastLoginTime')
    books = relationship('Book', protected=False)

    def _set_password(self, password):
        self._password = 'hashed:%s' % password

    def _get_password(self):  # pragma: no cover
        return self._password

    password = synonym(
        '_password',
        descriptor=property(_get_password, _set_password),
        protected=True
    )

    _avatar = Field('avatar', Unicode(255), nullable=True, protected=True)
    def _set_avatar(self, avatar):  # pragma: no cover
        self._avatar = 'avatar:%s' % avatar

    def _get_avatar(self):  # pragma: no cover
        return self._avatar

    avatar = synonym(
        '_avatar',
        descriptor=property(_get_avatar, _set_avatar),
        protected=False,
        json='avatarImage'
    )


class Root(JsonPatchControllerMixin, ModelRestController):
    __model__ = Member

    @json
    @commit
    def post(self):
        m = Member()
        m.keywords.append('Hello')
        m.update_from_request()
        DBSession.add(m)
        return m

    @json
    @Member.expose
    def get(self, title: str=None):
        query = DBSession.query(Member)
        if title:
            return query.filter(Member.title == title).one_or_none()
        return query

    @json
    @Member.expose
    def me(self):
        return dict(
            title='me'
        )


class TestBaseModel(ApplicableTestCase):
    __controller_factory__ = Root

    def test_update_from_request(self):
        with self.given(
                'Posting a form',
                verb='POST',
                form=dict(
                    title='test',
                    firstName='test',
                    lastName='test',
                    email='test@example.com',
                    password='123456',
                    birth='2001-01-01',
                    weight=1.1,
                    visible='false',
                    lastLoginTime='2017-10-10T15:44:30.000',
                    isActive=True
                )):
            assert response.json['title'] == 'test'
            assert 'avatar' not in response.json
            assert '_avatar' not in response.json
            assert 'avatarImage' in response.json

            # 400 for sending relationship attribute
            when(
                'Sending a relationship attribute',
                form=Update(email='test2@example.com', books=[])
            )
            assert status == '200 OK'
            assert {
                'Keywords': [{'id': 2, 'keyword': 'Hello'}],
                'birth': '2001-01-01',
                'books': [],
                'email': 'test2@example.com',
                'firstName': 'test',
                'fullName': 'test test',
                'id': 2,
                'lastName': 'test',
                'title': 'test',
                'weight': '1.1000000000'
            }.items() <= response.json.items()

    def test_pagination(self):
        with self.given(
                'Getting a single object using pagination',
                query=dict(take=1)
            ):

            assert len(response.json) == 1
            assert response.json[0]['title'] == 'test'
            assert response.json[0]['visible'] == False

            when(
                'Trying to get an non-existence db object',
                '/non-existence-user'
            )
            assert status == 404

            when('Getting a plain dictionary', '/me')
            assert response.json == {'title': 'me'}

    def test_iter_columns(self):
        columns = {
            c.key: c for c in Member.iter_columns(
                relationships=False,
                synonyms=False,
                composites=False
            )
        }
        assert len(columns) == 16
        assert 'name' not in columns
        assert 'password' not in columns
        assert '_password' in columns

    def test_iter_json_columns(self):
        columns = {
            c.key: c for c in Member.iter_json_columns(
                include_readonly_columns=False,
                include_protected_columns=False
            )
        }
        assert len(columns) == 12
        assert 'name' not in columns
        assert 'password' not in columns
        assert '_password' not in columns
        assert '_avatar' not in columns
        assert 'avatar' in columns

    def test_metadata(self):
        with self.given(
            'Fetching the metadata',
            verb='METADATA'
        ):
            assert 'fields' in response.json
            assert 'name' in response.json
            assert 'primaryKeys' in response.json
            assert 'id' in response.json['primaryKeys']
            assert response.json['name'] == 'Member'

            fields = response.json['fields']
            assert 'id' in fields
            assert 'firstName' in fields
            assert fields['id']['primaryKey'] == True
            assert fields['email']['primaryKey'] == False
            assert fields['title']['primaryKey'] == False
            assert fields['title']['minLength'] == 2
            assert fields['title']['maxLength'] == 50
            assert fields['email']['pattern'] == \
                '(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)'

            assert fields['firstName']['name'] == 'firstName'
            assert fields['firstName']['key'] == 'first_name'
            assert fields['weight']['default'] == 50
            assert fields['visible']['not_none'] == None
            assert fields['email']['watermark'] == 'Email'
            assert fields['email']['label'] == 'Email'
            assert fields['email']['example'] == 'user@example.com'

